"""
Contains a collection of strings that appear many times in the fruitbat
documentation.
"""


def _docstr_sub(*args, **kwargs):
    if args and kwargs:
        raise InputError("Only positional or keyword arguments are allowed")
    else:
        params = args or kwargs

    def do_sub(docstr):
        if docstr.__doc__:
            docstr.__doc__ = docstr.__doc__ % params
        else:
            raise InputError("Function has no docstring for substitution")

        return(docstr)
    return(do_sub)


_methods_doc = "``'batten2019'``, ``'zhang2018'``, ``'inoue2004'``, ``'ioka2003'``"

_cosmo_doc = "``'wmap2013'``, ``planck2013``, ``'planck2015'``, ``planck2018``, ``eagle``"

_dm_units_doc = ":math:`\\rm{pc\\ cm^{-3}}`"



from fruitbat import Frb
from astropy.coordinates import SkyCoord
from astropy import units as u
import pytest


class Test_Frb_Class(object):

    # Create FRB objects for testing
    frb = Frb("Utmost1", dm=1000, dm_uncert=0.0)
    frb_raj_decj = Frb('Utmost1', dm=1000, raj="11:05:50.0", decj="-8:34:12.0")
    frb_gl_gb = Frb('Utmost1', dm=1000, gl="30.5", gb="-60.2") 

    # Test that methods returns the correct value for DM=1000 and planck2018
    def test_methods(self):
        methods = {
                   "ioka2003" : 0.808948679473659,
                   "inoue2004": 0.9838829149238645,
                   "zhang2018": 1.1092684927893448,
                }
    

        for method in methods.keys():
            z = self.frb.calc_redshift(method=method, cosmology="planck2018")
            assert z == methods[method], "Test failed for: {}".format(method)
   

    def test_frb_calc_skycoords_raj_decj(self):
        ra_str = "11:05:50.0"
        dec_str = "-8:34:12.0"
    
        skycoords = self.frb_raj_decj.calc_skycoords()
        test_skycoords = SkyCoord(ra_str, dec_str, frame="icrs", 
                                   unit=(u.hourangle, u.deg))

        ra, dec = skycoords.ra.value, skycoords.dec.value
        test_ra, test_dec = test_skycoords.ra.value, test_skycoords.dec.value
        assert (ra, dec) == (test_ra, test_dec)
    
    
    def test_frb_calc_skycoords_gl_gb(self):
        gl_str = "30.5"
        gb_str = "-60.2"
    
        skycoords = self.frb_gl_gb.calc_skycoords()
        test_skycoords = SkyCoord(gl_str, gb_str, frame="galactic", unit=u.deg)

        gl, gb = skycoords.l.value, skycoords.b.value
        test_gl, test_gb = test_skycoords.l.value, test_skycoords.b.value
        assert (gl, gb) == (test_gl, test_gb)
    
    
    def test_invalid_method(self):
        invalid_method = "jacqui1992" 
        with pytest.raises(ValueError):
            self.frb.calc_redshift(invalid_method)
            
    
    def test_invalid_cosmology(self):
        invalid_cosmology = "cosmos_1964"
        with pytest.raises(ValueError):
            self.frb.calc_redshift(invalid_cosmology)



"""
Utility functions for Fruitbat
"""

import os

import numpy as np
import scipy.integrate as integrate
import scipy.interpolate as interpolate
import astropy.constants as CONST
import astropy.units as u

def load_lookup_table(filename, data_dir='data'):
    """
    Opens a saved `.npy` file containing an interpolated 1D function.

    Parameters
    ----------
    filename : str
        The name of the file to load.

    data_dir : str, optional
        The directory containing the data. The whole path must be specified
        except if :attr:`data_dir` == 'data' then it will search in the
        `data` subdirectory of the source code. Default: 'data'

    Returns
    -------
    scipy.interpolate.interpolate.interp1d
        Function
    """
    if data_dir == 'data':
        data_dir = os.path.join(os.path.dirname(__file__), 'data')
    file = os.path.join(data_dir, filename)
    return np.load(file)[()]


def create_lookup_table(filename, method, cosmology, zmin=0, zmax=30,
                        num_samples=1e5, *args, **kwargs):
    """
    Creates a lookup table

    Parameters
    ----------
    filename : str

    method : str

    cosmology : dict
        A dictionary containing the cosmology parameters for the Hubble 
        constant, Matter density, Baryon density and Dark Energy density. 
        `cosmology` must contain values for the following keys: 
        ``'H0'``, ``'Omega_m'``, ``'Omega_b'``, ``'Omega_L'``

    zmin : int or float, optional
        The minimum redshift in the table. Default: 0

    zmax : int or float, optional
        The maximum redshift in the table. Default: 30

    num_samples : int, optional
        The number of dispersion measure samples to perform before 
        interpolation. Default: 100000

    Returns
    -------
    None
    """
    
    method_table_dict = {"inoue2004": _create_lookup_table_inoue2004,
                         "zhang2018": _create_lookup_table_zhang2018,
                         "ioka2003": _create_lookup_table_ioka2003
                        }

    
    method_table_dict[method](filename, cosmology, zmin=zmin, zmax=zmax, 
                              num_samples=num_samples, args=args, 
                              kwargs=kwargs)




    
def _fz_integrand(z, cosmology):
    """
    Calculate the integrand for a given redshift and cosmology. This integral
    appears in zhang2018, inoue2004 and ioka2003. 

    Parameters
    ----------
    z : float
        The redshift to evaluate the integral

    cosmology
    """
    top = 1 + z
    bot = cosmology["Omega_m"] * (1 + z)**3 + cosmology["Omega_L"]

    return top / np.sqrt(bot)
 

def _create_lookup_table_ioka2003(filename, cosmo, zmin=0, zmax=30, 
                                  num_samples=1e5):
    """
    Creates an interpolated 1D DM-z look up table using the Ioka (2003)
    relation and a given cosmology.

    Parameters
    ----------
    filename: str

    cosmo: dict

    zmin: int or float, optional
        The minimum redshift for the table. The output table will
        not be able to estimate redshifts lower than this value. Default: 0
    
    zmax: int or float, optional
        The maximum redshift for the table. The output table will
        not be able to estimate redshifts higher than this value. Default: 30

    num_samples: int, optional
        Default: 100000

    Returns
    -------
    None
    """
        
    def _calc_dm(z, cosmo):
        """
        Calculate the dispersion measure from a redshift given a cosmology
        using the Ioka (2003) relation.
        """
        # Check that the user has provided all the required values
        cosmo_required_keys = ["HO", "Omega_m", "Omega_b", "Omega_L"]
        _check_keys_in_dict(cosmo, cosmo_required_keys)

        coeff_top = 3 * CONST.c * cosmo["HO"] * cosmo["Omega_b"]
        coeff_bot = 8 * np.pi * CONST.G * CONST.m_p
        coeff = coeff_top / coeff_bot
        coeff = coeff.to("pc cm**-3")

        dm = coeff * integrate.quad(_fz_integrand, 0, z, args=(cosmo))[0]

        return dm.value

    interp = _perform_interpolation(dm_func=_calc_dm, cosmology=cosmo, 
                                    zmin=zmin, zmax=zmax, 
                                    num_samples=num_samples)    
    
    _save_lookup_table(interp, filename)





def _create_lookup_table_inoue2004(filename, cosmo, zmin=0, zmax=30, 
                                   num_samples=1e5, *args, **kwargs):
    """
    Creates an interpolated 1D DM-z look up table using the Inoue (2004)
    relation and a given cosmology.

    Parameters
    ----------
    filename: str

    cosmo: dict

    zmin: int or float, optional
        The minimum redshift for the table. The output table will
        not be able to estimate redshifts lower than this value. Default: 0
    
    zmax: int or float, optional
        The maximum redshift for the table. The output table will
        not be able to estimate redshifts higher than this value. Default: 30

    num_samples: int, optional
        Default: 100000

    Returns
    -------
    None
    """
        
    def _calc_dm(z, cosmo):
        """
        Calculate the dispersion measure from a redshift given a cosmology
        using the Inoue (2004) relation.
        """
        # Check that the user has provided all the required values
        cosmo_required_keys = ["HO", "Omega_m", "Omega_b", "Omega_L"]
        _check_keys_in_dict(cosmo, cosmo_required_keys)

        inoue_n_e_0 = 9.2e-10 * ((u.Mpc**2 * u.s**2) / (u.km**2 * u.cm**3)) 

        coeff = inoue_n_e_0 * CONST.c * cosmo["Omega_b"] * cosmo["HO"]
        coeff = coeff.to("pc cm**-3")

        dm = coeff * integrate.quad(_fz_integrand, 0, z, args=(cosmo))[0]

        return dm.value

    interp = _perform_interpolation(dm_func=_calc_dm, cosmology=cosmo, 
                                    zmin=zmin, zmax=zmax, 
                                    num_samples=num_samples, args=args, 
                                    kwargs=kwargs)    
    
    _save_lookup_table(interp, filename)


def _create_lookup_table_zhang2018(filename, cosmo, zmin=0, zmax=30, 
                                   num_samples=1e5, f_igm=0.83, 
                                   free_elec=0.875, *args, **kwargs):
    """
    Creates an interpolated 1D DM-z look up table using the Zhang (2018)
    relation and a given cosmology.

    Parameters
    ----------
    filename: str

    cosmo: dict

    zmin: int or float, optional
        Default: 0

    zmax: int or float, optional
        Default: 30

    num_samples: int, optional
        Default: 100000

    f_igm: float, optional
        Default: 0.83

    free_elec: float, optional
        Default: 0.875
    """

    def _calc_dm(z, cosmo, approx=False, *args, **kwargs):

        if approx:
            # If the user wants to create an approximate table similar to
            # Zhang 2018 without explicitly solving the integral.
            dm = 1168 * f_igm * free_elec * z
            return dm

        else:
            # Check that the user has provided all the required values
            cosmo_required_keys = ["HO", "Omega_m", "Omega_b", "Omega_L"]
            _check_keys_in_dict(cosmo, cosmo_required_keys)

            coeff_top = 3 * CONST.c * cosmo["HO"] * cosmo["Omega_b"]
            coeff_bot = 8 * np.pi * CONST.G * CONST.m_p
            coeff = coeff_top / coeff_bot

            coeff = coeff.to("pc cm**-3")
            dm = coeff * f_igm * free_elec * integrate.quad(_fz_integrand, 0, z,
                                                           args=(cosmo))[0]
            return dm.value

    interp = _perform_interpolation(dm_func=_calc_dm, cosmology=cosmo, 
                                    zmin=zmin, zmax=zmax, 
                                    num_samples=num_samples, args=args, 
                                    kwargs=kwargs)

    _save_lookup_table(interp, filename)

def _perform_interpolation(dm_func=None, zmin=None, zmax=None, cosmology=None,
                           num_samples=None, *args, **kwargs):
    """
    

    Parameters
    ----------

    Returns
    -------


    """

    z_vals = np.linspace(zmin, zmax, num_samples)
    dm_vals = np.array([dm_func(z, cosmology) for z in z_vals])
    interp = interpolate.interp1d(dm_vals, z_vals)

    return interp

def _check_keys_in_dict(dictionary, keys):
    """
    Checks that a list of keys exist in a dictionary.

    Parameters
    ----------
    dictionary: dict
        

    keys: list of strings
        The keys that the dictionary must have.

    Returns
    -------
    bool:
        True
    """
    if not all(key in dictionary for key in keys):
        raise KeyError("Dictionary missing key values." 
                       "Requires: {}".format(keys))
    return True


def _save_lookup_table(table, filename):
    """
    Saves the lookup table to a .npy file
    """
    np.save(filename, table)
    return None



from . import utils

# __all__ = ['dm_to_redshift']
from ._fruitbatstrings import (_docstr_sub, _cosmo_doc,
                               _methods_doc, _dm_units_doc)
from .cosmology import cosmology_keys

@_docstr_sub(dmunits=_dm_units_doc, methods=_methods_doc, cosmo=_cosmo_doc)
def redshift(dm, dm_uncert=0.0, method='inoue2004', cosmology='planck2018'):
    """
    Returns the redshift of a given dispersion measure using a
    specified DM-z relation.

    Parameters
    ----------
    dm : float
        Dispersion Measure. Units: %(dmunits)s

    dm_uncert : float or None
        The uncertainty in the dispersion measure. Units: %(dmunits)s

    method : string, optional
        The DM-z relation to use to calculate the redshift.
        Avaliable methods are: %(methods)s. Default: `'inoue2004'`

    cosmology : string, optional
        Avaliable cosmologies: %(cosmo)s. Default: `'planck2018'`

    Returns
    -------
    z : float
        Redshift

    z_err : float
        The uncertianty in the redshift estimation. If `dm_uncert` is `None`
        then `z_err` = 0.

    Notes
    -----

    Cosmology_ has a list of the cosmological parameters used in each
    cosmology method.

    .. _Cosmology: https://fruitbat.readthedocs.io/en/latest/cosmology.html
    """

    valid_methods = ['batten2019', 'zhang2018', 'inoue2004', 'ioka2003']

    if method not in valid_methods:
        raise ValueError("""Method '{m}' is not a valid method.
            Valid methods are: %(methods_doc)s""")

    if cosmology not in cosmology_keys():
        raise ValueError("""Cosmology '{c}' is not a valid cosmology.
            Valid cosmologies are: %(cosmo)s""")

    z = _get_redshift_from_table(dm, method, cosmology)

    return z


#def _redshift_batten2019(dm, dm_uncert=0.0):
#    """
#    Calculates a redshft from a dispersion measure using the DM-z
#    relation from Batten, A. J. 2019, ....
#    """
#    return 12.0
#
#@_docstr_sub(dm_units=_dm_units_doc)
#def _redshift_inoue2004(dm, dm_uncert=0.0, cosmology="planck2018"):
#    """
#    Calculates a redshift from a dispersion measure using the DM-z
#    relation from Inoue2004_
#
#    Parameters
#    ----------
#    dm : float
#        Units: %(dm_units)s
#
#    dm_uncert : float, optional
#        Default: :0.0
#
#    cosmology : str, optional
#        Default: `'planck2018'`
#
#    Returns
#    -------
#    z : float
#        The redshift of the FRB.
#    z_uncert : tuple of floats
#        The uncertainty in the redshift of the FRB.
#
#    Notes
#    -----
#    Inoue2004_ presents the following dispersion measure and redshift relation.
#   
#    .. math ::
#        \\mathrm{DM}(z) =0.92\\times 10^{-5} \\Omega_b h^2 c \\int_0^z
#        \\frac{1 + z'}{H_0\\left(\\Omega_m (1 + z)^3 +
#        \\Omega_\\Lambda\\right)^{1/2}} dz'
#
#
#    .. _Inoue2004: http://adsabs.harvard.edu/abs/2004MNRAS.348..999I
#    """
#
#    cosmo_dict = {
#        "wmap2013": "inoue2004_wmap2013.npy",
#        "planck2013": "inoue2004_planck2013.npy",
#        "planck2015": "inoue2004_planck2015.npy", 
#        "planck2018": "inoue2004_planck2018.npy",
#        "eagle": "inoue2004_eagle.npy",
#    }
#
#    lookup_table = utils.load_lookup_table(cosmo_dict[cosmology])
#
#    z = _get_redshift_from_table(lookup_table, dm, dm_uncert) 
#    #z, dz_low, dz_high = _get_redshift_from_table(lookup_table, dm, dm_uncert) 
#    return z#, (dz_low, dz_high)
#
#
#def _redshift_ioka2003(dm, dm_uncert=0.0):
#    """
#    Calculates a redshift from a dispersion measure using the DM-z
#    relation from Ioka, K. 2003, ApJL, 598, L79
#    """
#    cosmo_dict = {
#        "wmap2013": "ioka2003_wmap2013.npy",
#        "planck2013": "ioka2003_planck2013.npy",
#        "planck2015": "ioka2003_planck2015.npy",
#        "planck2018": "ioka2003_planck2018.npy",
#        "eagle": "ioka2003_eagle.npy"
#        }
#    lookup_table = utils.load_lookup_table(cosmo_dict[cosmology])
#    z = _get_redshift_from_table(lookup_table, dm, dm_uncert)
#
#    return z
#
#
#def _redshift_zhang2018(dm, dm_uncert=0.0, cosmology="planck2018"):
#    """
#    Calculates a redshift from a dispersion measure using the DM-z
#    relation from Zhang2018_
#
#
#    .. _Zhang2018: https://ui.adsabs.harvard.edu/#abs/arXiv:1808.05277
#    """
#
#    cosmo_dict = {
#        "wmap2013": "zhang2018_wmap2013.npy",
#        "planck2013": "zhang2018_planck2013.npy",
#        "planck2015": "zhang2018_planck2015.npy",
#        "planck2018": "zhang2018_planck2018.npy",
#        "eagle": "zhang2018_eagle.npy"
#        }
#    lookup_table = utils.load_lookup_table(cosmo_dict[cosmology])
#    z = _get_redshift_from_table(lookup_table, dm, dm_uncert)
#
#    return z
#

def _get_redshift_from_table(dm, method, cosmology):
    """
    Loads a lookup table for a corresponding method and cosmology and reads
    the value of the redshift correspoding the the DM value.

    Parameters
    ----------
    dm : int or float

    method: str

    cosmology: str
        The cosmology keyword 

    Returns
    -------
    float
        The redshift corresponding the the DM using the provided method and 
        cosmology.
    """

    table_name = "_".join([method, cosmology]) 
    table_extension = "npy"
    filename = ".".join([table_name, table_extension])
    lookup_table = utils.load_lookup_table(filename)
    z = lookup_table(dm)[()]
    return z



"""
Provides definition of the Frb class and member functions.
"""

from astropy.coordinates import SkyCoord
import astropy.units as u

from . import estimate
from ._fruitbatstrings import (_docstr_sub, _methods_doc, 
                               _cosmo_doc, _dm_units_doc)

@_docstr_sub(dm_units=_dm_units_doc)
class Frb(object):
    """
    Defines the :class:`~Frb` class in the **fruitbat** package.

    Attributes
    ----------
    name : str
        The name of the Frb object.

    dm : float
        The observed dispersion measure of the FRB without Milky Way
        subtraction.
        Units: %(dm_units)s

    dm_uncert : float, optional
        The uncertainty in the dispersion measure.
        Units: %(dm_units)s Default: 0.0

    dm_galaxy : float, optional
        The modelled contribution to the FRB DM by electrons in the Milky Way.
        Units: %(dm_units)s Default: 0.0

    dm_host : float, optional
        The dispersion measure of the FRB host galaxy.
        Units: %(dm_units)s Default: 0.0

    dm_excess : float or None, optional
        The DM excess of the FRB over the estimated Galactic DM. If
        `dm_excess` is *None*, then `dm_excess` is calculated by
        :math:`\\rm{DM - DM_{galaxy}}`
        Units: %(dm_units)s Default: *None*

    dm_index : float or None, optional
        The dispersion measure index of the burst :math:`\\alpha` such that
        :math:`\\rm{DM} \\propto \\nu^{-\\alpha}` Default: *None*

    z : float or None, optional
        The redshift of the burst. This assumes that the entire `dm_excess`
        arrives from the IGM and the host galaxy of the FRB and any
        surrounding material contribute nothing to the total DM. This
        should be taken as an upper limit to the bursts true redshift.
        Default: *None*

    z_uncert : float, optional
        The uncertainty in the redshift of the FRB.
        Default: 0.0

    scatt_index : float or None, optional
        The scattering index (:math:`\\beta`) of the FRB pulse. The
        scattering index describes how the width (:math:`\\rm{W}`) of the
        FRB pulse evolves with frequency :math:`\\nu` such that
        :math:`\\rm{W} \\propto \\nu^{-\\beta}` Default: *None*

    snr : float or None, optional
        The signal-to-noise of the burst.
        Default: *None*

    w_obs : float or None, optional
        The observed width of the pulse obtained by a pulse fitting algorithm.
        Units: :math:`\\rm{ms}` Default: *None*

    s_peak_obs : float or None, optional
        The observed peak flux density of the burst.
        Units: :math:`\\rm{Jy}` Default: *None*

    f_obs : float or None, optional
        The observed fluence of the FRB. If `f_obs` is *None* and both
        `w_obs` and `s_peak_obs` are not *None* then `f_obs` is
        calculated by :math:`\\rm{W_{obs} \\times S_{peak,obs}}`
        Units: :math:`\\rm{Jy\\ ms}` Default: *None*

    raj and decj : str or None, optional
        The right ascension and declination in J2000 coordinates of the 
        pointing centre of the detection beam. This corresponds only to the 
        positioning of
        the beam centre. Default: *None*

    gl and gb : str or None, optional
        The Galactic longitude and latitude in degree of the best estimate of 
        the best estimate of the FRB position. Default: *None*

    skycoords : astropy.coordinates.sky_coordinate.SkyCoord or None, optional
        The skycoords of the FRB. This is calculated from either (raj, decj) or
        (gl, gb)
    """

    def __init__(self, name, dm, dm_uncert=0.0, dm_galaxy=0.0, dm_host=0.0, 
                 dm_excess=None, dm_index=None, z=None, z_uncert=0.0, 
                 scatt_index=None, snr=None, w_obs=None, s_peak_obs=None, 
                 f_obs=None, raj=None, decj=None, gl=None, gb=None, 
                 skycoords=None):

        self.name = name
        self.dm = dm
        self.dm_uncert = dm_uncert
        self.dm_galaxy = dm_galaxy
        self.dm_host = dm_host

        # Calculate dm_excess from existing parameters if it is not given.
        if not dm_excess:
            self.calc_dm_excess()
        else:
            self.dm_excess = dm_excess

        self.dm_index = dm_index
        self.z = z
        self.z_uncert = z_uncert
        self.scatt_index = scatt_index
        self.snr = snr
        self.w_obs = w_obs
        self.s_peak_obs = s_peak_obs
        self.raj = raj
        self.decj = decj
        self.gl = gl
        self.gb = gb

        # Calculate F_obs if s_peak_obs and w_obs are given
        if (not f_obs) and (s_peak_obs and w_obs):
            self.calc_f_obs()
        else:
            self.f_obs = f_obs

        # Calculate the skycoords of the FRB from (raj, decj) or (gl, gb)
        if skycoords:
            self.skycoords = skycoords
        elif (raj and decj) is not None or (gl and gb) is not None:
            self.skycoords = self.calc_skycoords()



        if skycoords is not None:
            self.skycoords = skycoords
        elif raj is not None and decj is not None:
            self.skycoords = SkyCoord(raj, decj, frame="icrs", 
                                      unit=(u.hourangle, u.deg))
        elif gl is not None and gb is not None:
            self.skycoords = SkyCoord(gl, gb, frame="galactic", unit=u.deg)


    def __repr__(self):
        return 'Frb({0})'.format(vars(self))

    @_docstr_sub(methods=_methods_doc, cosmo=_cosmo_doc)
    def calc_redshift(self, method='inoue2004', cosmology="planck2018"):
        """
        Calculate the redshift of the FRB from its dm or dm_excess

        Parameters
        ----------
        method : str, optional
            The approximation to use when calculating the redshift.
            Avaliable methods:  %(methods)s

        cosmology : str, optional
            The method `inoue2004` has the option to choose which cosmology
            to assume when performing the redshift estimation.
            Avaliable cosmologies: %(cosmo)s

        Returns
        -------
        z : float
            The redshift of the FRB.

        z_uncert : float
            The uncertainty of the redshift estimation.


        Notes
        -----
        The methods_ section in the documentation has a discription of each 
        methods and where they should apply.

        The cosmology_ section of the documentation has a list of the 
        cosmological parameters used in each cosmology method.

        .. _cosmology: https://fruitbat.readthedocs.io/en/latest/user_guide/method_and_cosmology.html#cosmology
        .. _methods: https://fruitbat.readthedocs.io/en/latest/user_guide/method_and_cosmology.html#methods
        """

        z  = estimate.redshift(self.dm_excess, self.dm_uncert, 
                               method, cosmology)

        self.z = z
        #self.z_uncert = z_uncert

        return z


    def calc_skycoords(self):
        """
        Calculates the skycoord position of the FRB from (raj, decj) or (gl, gb)

        Returns
        -------
        skycoords : astropy.coordinates.sky_coordinate.SkyCoord
            The sky coordinates of the FRB.

        """

        if self.raj is not None and self.decj is not None:
            skycoords = SkyCoord(self.raj, self.decj, frame="icrs", 
                                 unit=(u.hourangle, u.deg))
        elif self.gl is not None and self.gb is not None:
            skycoords = SkyCoord(self.gl, self.gb, frame="galactic", 
                                unit=u.deg)
        else:
            raise ValueError("To calculate skycoords either (`raj` and `decj`)"
                             "or (`gl`, `gb`) must be provided")


        return skycoords


    def calc_dm_excess(self):
        """
        Calculates the dispersion measure excess of the FRB by subtracting
        the DM contribution from the Milky Way.

        Returns
        -------
        dm_excess : float
            The dispersion measure excess.

        Notes
        -----
        :math:`\\rm{DM_{excess}}` is calculated as follows:

        .. math::

            DM_{excess} = DM - DM_{galaxy}
        """
        dm_excess = self.dm - self.dm_galaxy
        self.dm_excess = dm_excess
        return dm_excess


    def calc_dm_igm(self):
        """
        Calculates the dispersion measure of the intergalactic medium of the
        FRB. This can only be done is the dispersion measure and redshift of 
        the FRB host is known.

        Returns
        -------
        dm_igm : float
            The dispersion measure from the IGM.

        Notes
        -----
        :math:`DM_{IGM}` is calculated as follows:

        .. math::

           DM_{IGM} = DM_{excess} - \\frac{DM_{host}}{1 + z} 
        """

        if self.z is None:
            err_msg = ("z is None. Provide a non zero value for the " 
                       "FRB host redshift")
            raise ValueError(err_msg)

        if self.dm_host == 0.0:
            err_msg = ("dm_host = 0. The dm_igm will be the same as "
                       "dm_excess. Provide a non-zero value for dm_host")
            raise ValueError(err_msg)

        dm_igm = self.dm_excess - (self.dm_host / (1 + self.z))
        self.dm_igm = dm_igm
        return dm_igm



    def calc_f_obs(self):
        """
        Calculates the observed fluence of the FRB. This requires `w_obs`
        and `s_peak_obs` to not be *None*.

        Returns
        -------
        float
            The fluence of the FRB

        Notes
        -----
        :math:`\\rm{F_{obs}}` is calculated as follows:

        .. math::

            \\rm{F_{obs} = W_{obs} \\times S_{peak, obs}}
        """

        if (not self.w_obs) or (not self.s_peak_obs):
            err_msg = ("calc_f_obs requires both w_obs and s_peak_obs "
                       "to not be None")
            raise ValueError(err_msg)

        f_obs = self.w_obs * self.s_peak_obs
        self.f_obs = f_obs
        return f_obs



"""
Module for defining different cosmologies
"""
from astropy import units as u


def avaliable_cosmologies():
    """
    Returns
    -------
    dict
        A dictionary containing the parameters for each cosmology.
    """
    cosmology_dict = {
        "wmap2013": wmap2013(),
        "planck2013": planck2013(),
        "planck2015": planck2015(),
        "planck2018": planck2018(),
        "eagle": eagle()
    }

    return cosmology_dict 


def cosmology_keys():
    """
    Returns
    -------
    list
        A list containing the keyword for all avaliable cosmologies.
    """

    return avaliable_cosmologies().keys()


def planck2018():
    """

    https://ui.adsabs.harvard.edu/#abs/arXiv:1807.06209

    Returns
    -------
    dict
        Planck 2018 best fit with cmb lensing and bao
    """
    
    planck2018 = {}

    planck2018["HO"] = 67.66 * u.km / u.s / u.Mpc
    planck2018["Omega_m"] = 0.3111
    planck2018["Omega_L"] = 0.6889
    planck2018["Omega_b"] = 0.04897
    planck2018["Omega_c"] = 0.2606
    planck2018["sigma8"] = 0.8102 * u.Mpc

    return planck2018


def planck2015():
    """
    Returns
    -------
    dict
        Planck 2015 best fit + CMB lensing + baryonic acoustic oscillations
    """
    
    planck2015 = {}

    planck2015["HO"] = 67.74 * u.km / u.s / u.Mpc 
    planck2015["Omega_m"] = 0.3089
    planck2015["Omega_L"] = 0.6911
    planck2015["Omega_b"] = 0.04860
    planck2015["Omega_c"] = 0.2589
    planck2015["sigma8"] = 0.8159 * u.Mpc

    return planck2015


def planck2013():
    """
    Returns
    -------
    dict
        Planck 2013 best fit + bao
    """
    
    planck2013 = {}

    planck2013["HO"] = 67.80 * u.km / u.s / u.Mpc 
    planck2013["Omega_m"] = 0.3063
    planck2013["Omega_L"] = 0.692
    planck2013["Omega_b"] = 0.04816
    planck2013["Omega_c"] = 0.2582
    planck2013["sigma8"] = 0.826 * u.Mpc

    return planck2013


def wmap2013():
    """
    Returns
    -------
    dict
        wmap 2013 best fit final data release
    """
    
    wmap2013 = {}

    wmap2013["HO"] = 69.32 * u.km / u.s / u.Mpc 
    wmap2013["Omega_m"] = 0.2865
    wmap2013["Omega_L"] = 0.7135
    wmap2013["Omega_b"] = 0.04628
    wmap2013["Omega_c"] = 0.2402
    wmap2013["sigma8"] = 0.820 * u.Mpc

    return wmap2013

def eagle():
    """
    Returns
    -------
    dict
        The cosmology that is used in the EAGLE simulations
    """

    eagle = {}

    eagle["HO"] = 67.77 * u.km / u.s / u.Mpc
    eagle["Omega_m"] = 0.307
    eagle["Omega_L"] = 0.693
    eagle["Omega_b"] = 0.04825
    eagle["sigma8"] = 0.8288
    eagle["ns"] = 0.9611
    eagle["Y"] = 0.248

    return eagle



from setuptools import setup, find_packages
from codecs import open

with open("README.rst", "r") as f:
    long_description = f.read()

with open("requirements.txt", "r") as f:
    requirements = f.read().splitlines()


setup(
    name='fruitbat',
    version='0.0.2',
    author='Adam Batten',
    author_email='adamjbatten@gmail.com',
    url='https://github.com/abatten/fruitbat',
    download_url="https://pypi.org/project/fruitbat",
    project_urls={
        'Documentation': "https://fruitbat.readthedocs.io",
        'Source Code': "https://github.com/abatten/fruitbat"
        },
    description='Calculate the redshift of a FRB from its dispersion measure',
    long_description=long_description,
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: BSD License",
        "Operating System :: MacOS",
        "Operating System :: Unix",
        ],
    package_dir={"fruitbat": "fruitbat"},
    packages=find_packages(),
    keywords=("FRB redshift astronomy astrophysics fast radio burst"),
)
